import tf_geometric as tfg
import numpy as np
edge_index = [
    [0, 1, 2, 3, 0, 1, 2, 2, 1, 2, 3],
    [0, 1, 2, 3, 1, 2, 3, 3, 2, 1, 0]
]

edge_weight = [0.1, 0.2, 0.5, 0.2, 0.7, 0.8, 0.1, 0.2, 0.1, 0.5, 0.2]

graph0 = tfg.Graph(x=np.random.randn(10, 5), edge_index=edge_index, edge_weight=edge_weight)

edge_index = [
    [0, 1, 3, 3, 4, 4, 7],
    [0, 1, 2, 3, 6, 6, 0]
]

edge_weight = [0.5, 0.2, 0.1, 0.5, 0.2, 0.5, 0.1]
# edge_graph_index = [0, 1, 5, 3, 1, 3, 0]
graph1 = tfg.Graph(x=np.random.randn(8, 5), edge_index=edge_index, edge_weight=edge_weight)

batch_graph = tfg.BatchGraph.from_graphs([graph0, graph1])

print(batch_graph.edge_index)
print(batch_graph.edge_weight)
print(batch_graph.edge_graph_index)
batch_graph.convert_edge_to_directed()

print("=========")

print(batch_graph.edge_index)
print(batch_graph.edge_weight)
print(batch_graph.edge_graph_index)

# print(convert_edge_to_upper(edge_index, [edge_weight, edge_graph_index], merge_modes=["sum", "max"]))

# print(convert_edge_to_directed(edge_index, [edge_weight, edge_graph_index], ["sum", "max"]))

def test(a):
    """

    :param a:
    :return:
    """
    pass


# coding=utf-8

from tf_geometric.data.graph import Graph
from tf_geometric.data.dataset import DownloadableDataset
from tf_geometric.utils.graph_utils import convert_edge_to_directed
from ogb_lite.nodeproppred import NodePropPredDataset
import numpy as np


class OGBNodePropPredDataset(DownloadableDataset):
    """
    OGB Node Property Prediction Datasets: https://ogb.stanford.edu/docs/nodeprop/
    """

    def __init__(self, dataset_name, dataset_root_path=None):
        """
        OGB Node Property Prediction Datasets: https://ogb.stanford.edu/docs/nodeprop/

        :param dataset_name: "ogbn-arxiv" | "ogbn-products" | "ogbn-proteins" | "ogbn-papers100M" | "ogbn-mag"
        :param dataset_root_path:
        """

        super().__init__(dataset_name=dataset_name,
                         download_urls=None,
                         download_file_name=None,
                         cache_name="cache.p",
                         dataset_root_path=dataset_root_path,
                         )

    # https://github.com/tkipf/gcn/blob/master/gcn/utils.py
    def process(self):
        dataset = NodePropPredDataset(name=self.dataset_name, root=self.download_root_path)

        graph, label = dataset[0]  # graph: library-agnostic graph object

        x = graph["node_feat"]
        edge_index = graph["edge_index"]

        # convert edge_index to directed
        edge_index, _ = convert_edge_to_directed(edge_index, None)

        label = label.flatten().astype(np.int32)
        graph = Graph(x=x, edge_index=edge_index, y=label)

        split_index = dataset.get_idx_split()
        train_index, valid_index, test_index = split_index["train"], split_index["valid"], split_index["test"]

        return graph, (train_index, valid_index, test_index)





# coding=utf-8

import warnings

warnings.simplefilter('always', DeprecationWarning)

import tensorflow as tf

if tf.__version__[0] == "1":
    tf.enable_eager_execution()

from . import nn, utils, data, datasets, layers, sparse
from tf_geometric.data.graph import Graph, BatchGraph
from tf_geometric.sparse.sparse_matrix import SparseMatrix
from tf_geometric.sparse.sparse_adj import SparseAdj



# coding=utf-8
import tensorflow as tf
from tf_geometric.nn.pool.common_pool import mean_pool, min_pool, max_pool, sum_pool


class CommonPool(tf.keras.Model):

    def __init__(self, pool_func, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pool_func = pool_func

    def call(self, inputs, training=None, mask=None):
        if len(inputs) == 2:
            x, node_graph_index = inputs
            num_graphs = None
        else:
            x, node_graph_index, num_graphs = inputs

        return self.pool_func(x, node_graph_index, num_graphs)


class MeanPool(tf.keras.Model):
    def __init__(self, *args, **kwargs):
        super().__init__(mean_pool, *args, **kwargs)


class MinPool(tf.keras.Model):
    def __init__(self, *args, **kwargs):
        super().__init__(min_pool, *args, **kwargs)


class MaxPool(tf.keras.Model):
    def __init__(self, *args, **kwargs):
        super().__init__(max_pool, *args, **kwargs)


class SumPool(tf.keras.Model):
    def __init__(self, *args, **kwargs):
        super().__init__(sum_pool, *args, **kwargs)


# coding=utf-8

import tensorflow as tf

from tf_geometric.nn.pool.sag_pool import sag_pool


class SAGPool(tf.keras.Model):
    """
    OOP API for SAGPool
    """

    def __init__(self, score_gnn, k=None, ratio=None, score_activation=None, *args, **kwargs):
        """
        SAGPool

        :param score_gnn: A GNN model to score nodes for the pooling, [x, edge_index, edge_weight] => node_score.
        :param k: Keep top k targets for each source
        :param ratio: Keep num_targets * ratio targets for each source
        :param score_activation: Activation to use for node_score before multiplying node_features with node_score
        """
        super().__init__(*args, **kwargs)
        self.score_gnn = score_gnn
        self.k = k
        self.ratio = ratio
        self.score_activation = score_activation

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight, node_graph_index]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Pooled graph: [pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index]
        """
        x, edge_index, edge_weight, node_graph_index = inputs

        return sag_pool(x, edge_index, edge_weight, node_graph_index, self.score_gnn,
                        k=self.k, ratio=self.ratio, score_activation=self.score_activation,
                        training=training, cache=cache)



# coding=utf-8
from scipy.sparse import csr_matrix
import numpy as np

from tf_geometric.datasets import MultiLabelBlogCatalogDataset

edge_index, y = MultiLabelBlogCatalogDataset().load_data()
num_nodes = y.shape[0]
num_classes = y.shape[1]

row, col = edge_index

adj = csr_matrix((np.ones_like(row), (row, col)), shape=[num_nodes, num_nodes])

adj_t = csr_matrix((np.ones_like(row), (col, row)), shape=[num_nodes, num_nodes])
import tensorflow as tf
import tf_geometric as tfg
x = tf.sparse.eye(num_nodes)
gcn = tfg.layers.GCN(num_classes)

# logits = gcn([x, edge_index])
#
# loss = tf.nn.sigmoid_cross_entropy_with_logits(
#     logits=logits,
#     labels=y
# )

graph = tfg.Graph(x=x, edge_index=edge_index, y=y)
logits = gcn([graph.x, graph.edge_index])

loss = tf.nn.sigmoid_cross_entropy_with_logits(
    logits=logits,
    labels=y
)

print(loss)



# print(edge_index)
# print(y)


# coding=utf-8

import tensorflow as tf
from tf_geometric.nn.sampling.drop_edge import drop_edge


class DropEdge(tf.keras.Model):
    def __init__(self, rate=0.5, force_undirected: bool = False):
        """
        DropEdge: Towards Deep Graph Convolutional Networks on Node Classification
        https://openreview.net/forum?id=Hkx1qkrKPr

        :param rate: dropout rate
        :param force_undirected: If set to `True`, will either
            drop or keep both edges of an undirected edge.
        """
        super().__init__()
        self.rate = rate
        self.force_undirected = force_undirected

        if self.rate < 0. or self.rate > 1.:
            raise ValueError('Dropout probability has to be between 0 and 1, '
                             'but got {}'.format(self.rate))

    def call(self, inputs, training=None, mask=None):
        return drop_edge(inputs=inputs, rate=self.rate,
                         force_undirected=self.force_undirected, training=training)



# coding=utf-8

import tensorflow as tf
import numpy as np

from tensorflow.python.ops.linalg.sparse.sparse_csr_matrix_ops import sparse_tensor_to_csr_sparse_matrix, \
    sparse_matrix_sparse_mat_mul, csr_sparse_matrix_to_sparse_tensor

from tf_geometric.utils.graph_utils import add_self_loop_edge, merge_duplicated_edge, remove_self_loop_edge

from tf_geometric.data.graph import Graph

from tf_geometric.nn.kernel.segment import segment_softmax
import warnings

"""
Sparse Matrix for Computation
"""


class SparseMatrix(object):
    # https://stackoverflow.com/questions/40252765/overriding-other-rmul-with-your-classs-mul
    __array_priority__ = 10000

    def __init__(self, edge_index, edge_weight=None, shape=None):
        """
        Sparse Matrix for efficient computation.
        :param edge_index:
        :param edge_weight:
        :param shape: [num_rows, num_cols], shape of the adjacency matrix.
        """

        self.edge_index = Graph.cast_edge_index(edge_index)

        edge_index_is_tensor = tf.is_tensor(edge_index)

        if edge_weight is not None:
            self.edge_weight = Graph.cast_edge_weight(edge_weight)
        else:
            if edge_index_is_tensor:
                num_edges = tf.shape(self.edge_index)[1]
                edge_weight = tf.ones([num_edges], dtype=tf.float32)
            else:
                num_edges = np.shape(self.edge_index)[1]
                edge_weight = np.ones([num_edges], dtype=np.float32)
            self.edge_weight = edge_weight

        self.edge_index, [self.edge_weight] = merge_duplicated_edge(self.edge_index, [self.edge_weight],
                                                                    merge_modes=["sum"])

        if shape is None:
            if edge_index_is_tensor:
                num_nodes = tf.reduce_max(edge_index) + 1
            else:
                num_nodes = np.max(edge_index) + 1
            self.shape = [num_nodes, num_nodes]
        else:
            self.shape = shape

    @property
    def row(self):
        return self.edge_index[0]

    @property
    def col(self):
        return self.edge_index[1]

    def add_self_loop(self, fill_weight=1.0):
        num_nodes = self.shape[0]
        updated_edge_index, updated_edge_weight = add_self_loop_edge(self.edge_index, num_nodes,
                                                                     edge_weight=self.edge_weight,
                                                                     fill_weight=fill_weight)
        return SparseMatrix(updated_edge_index, updated_edge_weight, self.shape)

    def remove_self_loop(self):
        updated_edge_index, updated_edge_weight = remove_self_loop_edge(self.edge_index, edge_weight=self.edge_weight)
        return SparseMatrix(updated_edge_index, updated_edge_weight, self.shape)

    def negative(self):
        return SparseMatrix(
            edge_index=self.edge_index,
            edge_weight=-self.edge_weight,
            shape=self.shape
        )

    def __neg__(self):
        return self.negative()

    def transpose(self):
        row, col = self.edge_index[0], self.edge_index[1]
        transposed_edge_index = tf.stack([col, row], axis=0)
        return SparseMatrix(transposed_edge_index, edge_weight=self.edge_weight, shape=[self.shape[1], self.shape[0]])

    def _reduce(self, segment_func, axis=-1, keepdims=False):

        # reduce by row
        if axis == -1 or axis == 1:
            reduce_axis = 0
        # reduce by col
        elif axis == 0 or axis == -2:
            reduce_axis = 1
        else:
            raise Exception("Invalid axis value: {}, axis shoud be -1, -2, 0, or 1".format(axis))

        reduce_index = self.edge_index[reduce_axis]
        num_reduced = self.shape[reduce_axis]

        reduced_edge_weight = segment_func(self.edge_weight, reduce_index, num_reduced)
        if keepdims:
            reduced_edge_weight = tf.expand_dims(reduced_edge_weight, axis=axis)
        return reduced_edge_weight

    def reduce_sum(self, axis=-1, keepdims=False):
        return self._reduce(tf.math.unsorted_segment_sum, axis=axis, keepdims=keepdims)

    def reduce_min(self, axis=-1, keepdims=False):
        return self._reduce(tf.math.unsorted_segment_min, axis=axis, keepdims=keepdims)

    def _matmul_dense(self, h):
        row, col = self.edge_index[0], self.edge_index[1]
        repeated_h = tf.gather(h, col)
        if self.edge_weight is not None:
            repeated_h *= tf.expand_dims(self.edge_weight, axis=-1)
        reduced_h = tf.math.unsorted_segment_sum(repeated_h, row, num_segments=self.shape[0])
        return reduced_h

    def _matmul_sparse(self, other):

        warnings.warn("The operation \"SparseMatrix @ SparseMatrix\" does not support gradient computation.")

        csr_matrix_a = self._to_csr_sparse_matrix()
        csr_matrix_b = other._to_csr_sparse_matrix()

        csr_matrix_c = sparse_matrix_sparse_mat_mul(
            a=csr_matrix_a, b=csr_matrix_b, type=self.edge_weight.dtype
        )

        sparse_tensor_c = csr_sparse_matrix_to_sparse_tensor(
            csr_matrix_c, type=self.edge_weight.dtype
        )

        return SparseMatrix.from_sparse_tensor(sparse_tensor_c)

    # sparse_adj @ other
    def matmul(self, other):
        if isinstance(other, SparseMatrix):
            return self._matmul_sparse(other)
        else:
            return self._matmul_dense(other)

    # h @ sparse_adj
    def rmatmul_dense(self, h):
        # h'
        transposed_h = tf.transpose(h, [1, 0])
        # sparse_adj' @ h'
        transpoed_output = self.transpose() @ transposed_h
        # h @ sparse_adj = (sparse_adj' @ h')'
        output = tf.transpose(transpoed_output, [1, 0])
        return output

    # # other_sparse_adj @ sparse_adj
    # def rmatmul_sparse(self, other):
    #     # h'
    #     transposed_other = other.transpose()
    #     # sparse_adj' @ h'
    #     transpoed_output = self.transpose() @ transposed_other
    #     # h @ sparse_adj = (sparse_adj' @ h')'
    #     output = transpoed_output.transpose()
    #     return output

    # self @ diagonal_matrix
    def matmul_diag(self, diagonal):
        col = self.edge_index[1]
        updated_edge_weight = self.edge_weight * tf.gather(diagonal, col)
        return SparseMatrix(self.edge_index, updated_edge_weight, self.shape)

    # self @ diagonal_matrix
    def rmatmul_diag(self, diagonal):
        row = self.edge_index[0]
        updated_edge_weight = tf.gather(diagonal, row) * self.edge_weight
        return SparseMatrix(self.edge_index, updated_edge_weight, self.shape)

    # self @ other (other is a dense tensor or SparseAdj)
    def __matmul__(self, other):
        return self.matmul(other)

    # h @ self (h is a dense tensor)
    def __rmatmul__(self, h):
        return self.rmatmul_dense(h)

    def eliminate_zeros(self):
        edge_index_is_tensor = tf.is_tensor(self.edge_index)
        edge_weight_is_tensor = tf.is_tensor(self.edge_weight)

        mask = tf.not_equal(self.edge_weight, 0.0)
        masked_edge_index = tf.boolean_mask(self.edge_index, mask, axis=1)
        masked_edge_weight = tf.boolean_mask(self.edge_weight, mask)

        if not edge_index_is_tensor:
            masked_edge_index = masked_edge_index.numpy()

        if not edge_weight_is_tensor:
            masked_edge_weight = masked_edge_weight.numpy()

        return SparseMatrix(masked_edge_index, masked_edge_weight, shape=self.shape)

    def merge(self, other_sparse_adj, merge_mode):
        """
        element-wise merge

        :param other_sparse_adj:
        :param merge_mode:
        :return:
        """

        edge_index_is_tensor = tf.is_tensor(self.edge_index)
        edge_weight_is_tensor = tf.is_tensor(self.edge_weight)

        combined_edge_index = tf.concat([self.edge_index, other_sparse_adj.edge_index], axis=1)
        combined_edge_weight = tf.concat([self.edge_weight, other_sparse_adj.edge_weight], axis=0)

        merged_edge_index, [merged_edge_weight] = merge_duplicated_edge(
            combined_edge_index, edge_props=[combined_edge_weight], merge_modes=[merge_mode])

        if not edge_index_is_tensor:
            merged_edge_index = merged_edge_index.numpy()

        if not edge_weight_is_tensor:
            merged_edge_weight = merged_edge_weight.numpy()

        merged_sparse_adj = SparseMatrix(merged_edge_index, merged_edge_weight, shape=self.shape)

        return merged_sparse_adj

    def __add__(self, other_sparse_adj):
        """
        element-wise sparse adj addition: self + other_sparse_adj
        :param other_sparse_adj:
        :return:
        """
        return self.merge(other_sparse_adj, merge_mode="sum")

    def __radd__(self, other_sparse_adj):
        """
        element-wise sparse adj addition: other_sparse_adj + self
        :param other_sparse_adj:
        :return:
        """
        return other_sparse_adj + self

    def __sub__(self, other):
        return self + other.negative()

    def __rsub__(self, other):
        return other - self

    def dropout(self, drop_rate, training=False):
        if training and drop_rate > 0.0:
            edge_weight = tf.compat.v2.nn.dropout(self.edge_weight, drop_rate)
        else:
            edge_weight = self.edge_weight
        return SparseMatrix(self.edge_index, edge_weight=edge_weight, shape=self.shape)

    def softmax(self, axis=-1):

        # reduce by row
        if axis == -1 or axis == 1:
            reduce_index = self.edge_index[0]
            num_reduced = self.shape[0]
        # reduce by col
        elif axis == 0 or axis == -2:
            reduce_index = self.edge_index[1]
            num_reduced = self.shape[1]
        else:
            raise Exception("Invalid axis value: {}, axis shoud be -1, -2, 0, or 1".format(axis))

        normed_edge_weight = segment_softmax(self.edge_weight, reduce_index, num_reduced)

        return SparseMatrix(self.edge_index, normed_edge_weight, shape=self.shape)

    def _to_csr_sparse_matrix(self):

        return sparse_tensor_to_csr_sparse_matrix(
            indices=tf.cast(tf.transpose(self.edge_index, [1, 0]), tf.int64),
            values=self.edge_weight,
            dense_shape=self.shape
        )

    @classmethod
    def from_diagonals(cls, diagonals):
        """
        Construct a SparseAdj from diagonals
        :param diagonals:
        :return:
        """
        num_nodes = tf.shape(diagonals)[0]
        row = tf.range(0, num_nodes, dtype=tf.int32)
        col = row
        edge_index = tf.stack([row, col], axis=0)
        return SparseMatrix(edge_index, diagonals, shape=[num_nodes, num_nodes])

    @classmethod
    def eye(cls, num_nodes):
        """
        Construct a SparseAdj with ones on diagonal
        :param diagonals:
        :return:
        """
        diagonals = tf.ones([num_nodes], dtype=tf.float32)
        return SparseMatrix.from_diagonals(diagonals)

    @classmethod
    def from_sparse_tensor(cls, sparse_tensor: tf.sparse.SparseTensor):
        return SparseMatrix(
            edge_index=tf.transpose(sparse_tensor.indices, [1, 0]),
            edge_weight=sparse_tensor.values,
            shape=sparse_tensor.dense_shape
        )

    def to_sparse_tensor(self):

        sparse_tensor = tf.sparse.SparseTensor(
            indices=tf.cast(tf.transpose(self.edge_index, [1, 0]), tf.int64),
            values=self.edge_weight,
            dense_shape=self.shape
        )
        sparse_tensor = tf.sparse.reorder(sparse_tensor)
        return sparse_tensor

    def to_dense(self):
        return tf.sparse.to_dense(self.to_sparse_tensor())

    def __str__(self):
        return "SparseAdj: \n" \
               "edge_index => \n" \
               "{}\n" \
               "edge_weight => {}\n" \
               "shape => {}".format(self.edge_index, self.edge_weight, self.shape)

    def __repr__(self):
        return self.__str__()



# coding=utf-8
import tensorflow as tf
import numpy as np

from tf_geometric.nn import mean_reducer, max_reducer, sum_reducer
from tf_geometric.nn.conv.gcn import gcn_mapper, gcn_norm_edge


def mean_graph_sage(x, edge_index, edge_weight,
                    self_kernel,
                    neighbor_kernel,
                    bias=None,
                    activation=None,
                    concat=True, normalize=False):
    """

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param self_kernel: Tensor, shape: [num_features, num_hidden_units], weight
    :param neighbor_kernel: Tensor, shape: [num_features, num_hidden_units], weight.
    :param bias: Tensor, shape: [num_output_features], bias
    :param activation: Activation function to use.
    :param normalize: If set to :obj:`True`, output features
                will be :math:`\ell_2`-normalized, *i.e.*,
                :math:`\frac{\mathbf{x}^{\prime}_i}
                {\| \mathbf{x}^{\prime}_i \|_2}`.
                (default: :obj:`False`)
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """
    num_nodes = tf.shape(x)[0]
    # num_edges = tf.shape(edge_index)[1]

    row, col = edge_index[0], edge_index[1]
    repeated_x = tf.gather(x, row)
    neighbor_x = tf.gather(x, col)

    if edge_weight is not None:
        neighbor_x = gcn_mapper(repeated_x, neighbor_x, edge_weight=edge_weight)

    neighbor_reduced_msg = mean_reducer(neighbor_x, row, num_nodes=num_nodes)

    neighbor_msg = neighbor_reduced_msg @ neighbor_kernel
    x = x @ self_kernel

    if concat:
        h = tf.concat([x, neighbor_msg], axis=1)
    else:
        h = x + neighbor_msg

    if bias is not None:
        h += bias

    if activation is not None:
        h = activation(h)

    if normalize:
        h = tf.nn.l2_normalize(h, axis=-1)

    return h



def sum_graph_sage(x, edge_index, edge_weight,
                    self_kernel,
                    neighbor_kernel,
                    bias=None,
                    activation=None,
                    concat=True, normalize=False):
    """

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param self_kernel: Tensor, shape: [num_features, num_hidden_units], weight
    :param neighbor_kernel: Tensor, shape: [num_features, num_hidden_units], weight.
    :param bias: Tensor, shape: [num_output_features], bias
    :param activation: Activation function to use.
    :param normalize: If set to :obj:`True`, output features
                will be :math:`\ell_2`-normalized, *i.e.*,
                :math:`\frac{\mathbf{x}^{\prime}_i}
                {\| \mathbf{x}^{\prime}_i \|_2}`.
                (default: :obj:`False`)
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """
    num_nodes = tf.shape(x)[0]
    # num_edges = tf.shape(edge_index)[1]

    row, col = edge_index[0], edge_index[1]
    repeated_x = tf.gather(x, row)
    neighbor_x = tf.gather(x, col)

    if edge_weight is not None:
        neighbor_x = gcn_mapper(repeated_x, neighbor_x, edge_weight=edge_weight)

    neighbor_reduced_msg = sum_reducer(neighbor_x, row, num_nodes=num_nodes)

    neighbor_msg = neighbor_reduced_msg @ neighbor_kernel
    x = x @ self_kernel

    if concat:
        h = tf.concat([x, neighbor_msg], axis=1)
    else:
        h = x + neighbor_msg

    if bias is not None:
        h += bias

    if activation is not None:
        h = activation(h)

    if normalize:
        h = tf.nn.l2_normalize(h, axis=-1)

    return h


def gcn_graph_sage(x, edge_index, edge_weight, kernel, bias=None, activation=None,
                   normalize=False, cache=None):
    """

        :param x: Tensor, shape: [num_nodes, num_features], node features
        :param edge_index: Tensor, shape: [2, num_edges], edge information
        :param edge_weight: Tensor or None, shape: [num_edges]
        :param kernel: Tensor, shape: [num_features, num_output_features], weight
        :param bias: Tensor, shape: [num_output_features], bias
        :param activation: Activation function to use.
        :param normalize: If set to :obj:`True`, output features
                will be :math:`\ell_2`-normalized, *i.e.*,
                :math:`\frac{\mathbf{x}^{\prime}_i}
                {\| \mathbf{x}^{\prime}_i \|_2}`.
                (default: :obj:`False`)
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """
    num_nodes = tf.shape(x)[0]
    num_edges = tf.shape(edge_index)[1]

    if edge_weight is not None:
        edge_weight = tf.ones([num_edges], dtype=tf.float32)

    updated_edge_index, normed_edge_weight = gcn_norm_edge(edge_index, num_nodes, edge_weight, cache)
    row, col = updated_edge_index
    repeated_x = tf.gather(x, row)
    neighbor_x = tf.gather(x, col)

    neighbor_x = gcn_mapper(repeated_x, neighbor_x, edge_weight=normed_edge_weight)

    reduced_msg = sum_reducer(neighbor_x, row, num_nodes=num_nodes)

    h = reduced_msg @ kernel
    if bias is not None:
        h += bias

    if activation is not None:
        h = activation(h)

    if normalize:
        h = tf.nn.l2_normalize(h, axis=-1)

    return h


def mean_pool_graph_sage(x, edge_index, edge_weight,
                         self_kernel, neighbor_mlp_kernel, neighbor_kernel,
                         neighbor_mlp_bias=None, bias=None, activation=None,
                         concat=True, normalize=False):
    """

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param self_kernel: Tensor, shape: [num_features, num_hidden_units], weight.
    :param neighbor_mlp_kernel: Tensor, shape: [num_features, num_hidden_units]. weight.
    :param neighbor_kernel: Tensor, shape: [num_hidden_units, num_hidden_units], weight.
    :param neighbor_mlp_bias: Tensor, shape: [num_hidden_units * 2], bias
    :param bias: Tensor, shape: [num_output_features], bias.
    :param activation: Activation function to use.
    :param normalize: If set to :obj:`True`, output features
                will be :math:`\ell_2`-normalized, *i.e.*,
                :math:`\frac{\mathbf{x}^{\prime}_i}
                {\| \mathbf{x}^{\prime}_i \|_2}`.
                (default: :obj:`False`)
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    num_nodes = tf.shape(x)[0]
    num_edges = tf.shape(edge_index)[1]

    if edge_weight is not None:
        edge_weight = tf.ones([num_edges], dtype=tf.float32)

    row, col = edge_index[0], edge_index[1]
    repeated_x = tf.gather(x, row)
    neighbor_x = tf.gather(x, col)

    neighbor_x = gcn_mapper(repeated_x, neighbor_x, edge_weight=edge_weight)

    h = neighbor_x @ neighbor_mlp_kernel
    if neighbor_mlp_bias is not None:
        h += neighbor_mlp_bias

    if activation is not None:
        h = activation(h)

    reduced_h = mean_reducer(h, row, num_nodes=num_nodes)

    from_neighbor = reduced_h @ neighbor_kernel
    from_x = x @ self_kernel

    if concat:
        output = tf.concat([from_x, from_neighbor], axis=1)
    else:
        output = from_x + from_neighbor

    if bias is not None:
        output += bias

    if activation is not None:
        output = activation(output)

    if normalize:
        output = tf.nn.l2_normalize(output, axis=-1)

    return output


def max_pool_graph_sage(x, edge_index, edge_weight,
                        self_kernel, neighbor_mlp_kernel, neighbor_kernel,
                        neighbor_mlp_bias=None, bias=None, activation=None,
                        concat=True, normalize=False):
    """

            :param x: Tensor, shape: [num_nodes, num_features], node features
            :param edge_index: Tensor, shape: [2, num_edges], edge information
            :param edge_weight: Tensor or None, shape: [num_edges]
            :param self_kernel: Tensor, shape: [num_features, num_hidden_units], weight.
            :param neighbor_mlp_kernel: Tensor, shape: [num_features, num_hidden_units]. weight.
            :param neighbor_kernel: Tensor, shape: [num_hidden_units, num_hidden_units], weight.
            :param neighbor_mlp_bias: Tensor, shape: [num_hidden_units * 2], bias
            :param bias: Tensor, shape: [num_output_features], bias.
            :param activation: Activation function to use.
            :param normalize: If set to :obj:`True`, output features
                        will be :math:`\ell_2`-normalized, *i.e.*,
                        :math:`\frac{\mathbf{x}^{\prime}_i}
                        {\| \mathbf{x}^{\prime}_i \|_2}`.
                        (default: :obj:`False`)
            :return: Updated node features (x), shape: [num_nodes, num_output_features]
            """
    num_nodes = tf.shape(x)[0]
    num_edges = tf.shape(edge_index)[1]

    if edge_weight is not None:
        edge_weight = tf.ones([num_edges], dtype=tf.float32)

    row, col = edge_index[0], edge_index[1]
    repeated_x = tf.gather(x, row)
    neighbor_x = tf.gather(x, col)

    neighbor_x = gcn_mapper(repeated_x, neighbor_x, edge_weight=edge_weight)

    h = neighbor_x @ neighbor_mlp_kernel
    if neighbor_mlp_bias is not None:
        h += neighbor_mlp_bias

    if activation is not None:
        h = activation(h)

    reduced_h = max_reducer(h, row, num_nodes=num_nodes)
    from_neighs = reduced_h @ neighbor_kernel
    from_x = x @ self_kernel

    if concat:
        output = tf.concat([from_x, from_neighs], axis=1)
    else:
        output = from_x + from_neighs

    if bias is not None:
        output += bias

    if activation is not None:
        output = activation(output)

    if normalize:
        output = tf.nn.l2_normalize(output, axis=-1)

    return output


def lstm_graph_sage(x, edge_index, lstm, self_kernel, neighbor_kernel,
                    bias=None, activation=None,
                    concat=True, normalize=False, training=False):
    """


    :param x: Tensor, shape: [num_nodes, num_features], node features.
    :param edge_index: Tensor, shape: [2, num_edges], edge information.
    :param lstm: Long Short-Term Merory.
    :param self_kernel: Tensor, shape: [num_features, num_hidden_units], weight.
    :param neighbor_kernel: Tensor, shape: [num_hidden_units, num_hidden_units], weight.
    :param bias: Tensor, shape: [num_output_features], bias.
    :param activation: Activation function to use.
    :param normalize: If set to :obj:`True`, output features
                will be :math:`\ell_2`-normalized, *i.e.*,
                :math:`\frac{\mathbf{x}^{\prime}_i}
                {\| \mathbf{x}^{\prime}_i \|_2}`.
                (default: :obj:`False`)
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """
    num_nodes = tf.shape(x)[0]
    num_edges = tf.shape(edge_index)[1]

    row, col = edge_index[0], edge_index[1]

    sort_index = tf.argsort(row, axis=-1, direction='ASCENDING')
    row = tf.gather(row, sort_index)
    col = tf.gather(col, sort_index)

    degree = tf.math.unsorted_segment_sum(tf.ones([num_edges], dtype=tf.int32), row, num_nodes)

    num_sampled_neighbors = tf.reduce_max(degree)
    num_edges_before_row = tf.concat([tf.zeros([1], dtype=tf.int32), tf.math.cumsum(degree)[:-1]], axis=0)
    matrix_col_index = tf.range(num_edges) - tf.gather(num_edges_before_row, row)

    neighbor_matrix = tf.fill((num_nodes, num_sampled_neighbors), num_nodes)
    neighbor_matrix = tf.tensor_scatter_nd_update(
        neighbor_matrix,
        tf.stack([row, matrix_col_index], axis=1),
        col
    )

    padded_x = tf.concat([x, tf.zeros([1, tf.shape(x)[-1]], x.dtype)], axis=0)

    neighbor_x = tf.gather(padded_x, neighbor_matrix)
    neighbor_h = lstm(neighbor_x, training=training)

    reduced_h = tf.reduce_mean(neighbor_h, axis=1)

    from_neighbor = reduced_h @ neighbor_kernel
    from_x = x @ self_kernel

    if concat:
        output = tf.concat([from_x, from_neighbor], axis=1)
    else:
        output = from_x + from_neighbor

    if bias is not None:
        output += bias

    if activation is not None:
        output = activation(output)

    if normalize:
        output = tf.nn.l2_normalize(output, axis=-1)

    return output



# coding=utf-8
import os

from scipy.sparse import csr_matrix
import numpy as np
os.environ["CUDA_VISIBLE_DEVICES"] = "6"

from tf_geometric import SparseMatrix
import tensorflow as tf



edge_index = [
    [0, 0, 0, 0, 1, 1, 4, 6],
    [0, 2, 4, 6, 2, 3, 6, 8]
]

adj = SparseMatrix(edge_index)
# print(adj)
# print("==========")
# print(adj.to_sparse_tensor())
# print(adj.to_dense())

# print(adj @ np.random.randn(9, 100).astype(np.float32))
# print(np.random.randn(100, 9).astype(np.float32) @ adj)
print(tf.convert_to_tensor(np.random.randn(100, 9).astype(np.float32)) @ adj)
# exit(0)


print(adj.to_dense())
print((adj @ adj).to_dense())

print("scipy")

m = csr_matrix(adj.to_dense().numpy())

print(m.todense())
print((m @ m).todense())

a = (adj @ adj).to_dense()
b = (m @ m).todense()

print("diff")
print(a.numpy() - b)



# print("1 ======")
# print(adj.to_dense())
# print("2 ======")
# print((adj - adj).eliminate_zeros().to_dense())
# print("3 ======")
# print((adj - adj.transpose()).to_dense())
#
# c = (adj - adj).eliminate_zeros()
# print(c)

# print(adj.reduce_sum(axis=-1, keepdims=True))
# h = np.random.randn(9, 20).astype(np.float32)
#
# print(adj @ h)
# print(adj.softmax(axis=-1))


# coding=utf-8
import tensorflow as tf

from tf_geometric.sparse import SparseMatrix
from tf_geometric.utils.graph_utils import convert_dense_adj_to_edge, remove_self_loop_edge, add_self_loop_edge


def cluster_pool(x, edge_index, edge_weight, assign_edge_index, assign_edge_weight, num_clusters, num_nodes=None):
    """
    Coarsen the input Graph based on cluster assignment of nodes and output pooled Graph.

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param assign_edge_index: Tensor, shape: [2, num_nodes], edge between clusters and nodes, where each edge
        denotes a node belongs to a specific cluster.
    :param assign_edge_weight: Tensor or None, shape: [num_nodes], the corresponding weight for assign_edge_index
    :param num_clusters: Number of clusters.
    :param num_nodes: Number of nodes, Optional, used for boosting performance.
    :return: Pooled Graph: [pooled_x, pooled_edge_index, pooled_edge_weight]
    """

    # Manually passing in num_nodes, num_clusters, and num_graphs can boost the performance
    if num_nodes is None:
        if x is None:
            raise Exception("Please provide num_nodes if x is None")
        else:
            num_nodes = tf.shape(x)[0]

    # compute S'AS
    S = SparseMatrix(assign_edge_index, assign_edge_weight, shape=[num_nodes, num_clusters])
    ST = S.transpose()
    A = SparseMatrix(edge_index, edge_weight, shape=[num_nodes, num_nodes]).to_dense()

    pooled_adj = ST @ A @ S

    pooled_edge_index, pooled_edge_weight = convert_dense_adj_to_edge(pooled_adj)

    if x is not None:
        # S'X
        pooled_x = ST @ x
    else:
        pooled_x = None

    return pooled_x, pooled_edge_index, pooled_edge_weight

# def cluster_pool_bak(x, edge_index, edge_weight, assign_edge_index, assign_edge_weight, num_clusters, num_nodes=None):
#     """
#     Coarsen the input Graph based on cluster assignment of nodes and output pooled Graph.
#
#     :param x: Tensor, shape: [num_nodes, num_features], node features
#     :param edge_index: Tensor, shape: [2, num_edges], edge information
#     :param edge_weight: Tensor or None, shape: [num_edges]
#     :param assign_edge_index: Tensor, shape: [2, num_nodes], edge between clusters and nodes, where each edge
#         denotes a node belongs to a specific cluster.
#     :param assign_edge_weight: Tensor or None, shape: [num_nodes], the corresponding weight for assign_edge_index
#     :param num_clusters: Number of clusters.
#     :param num_nodes: Number of nodes, Optional, used for boosting performance.
#     :return: Pooled Graph: [pooled_x, pooled_edge_index, pooled_edge_weight]
#     """
#
#     # Manually passing in num_nodes, num_clusters, and num_graphs can boost the performance
#     if num_nodes is None:
#         if x is None:
#             raise Exception("Please provide num_nodes if x is None")
#         else:
#             num_nodes = tf.shape(x)[0]
#
#     assign_row, assign_col = assign_edge_index[0], assign_edge_index[1]
#
#     transposed_sparse_assign_probs = tf.SparseTensor(
#         indices=tf.cast(tf.stack([assign_col, assign_row], axis=1), dtype=tf.int64),
#         values=assign_edge_weight,
#         dense_shape=[num_clusters, num_nodes]
#     )
#     transposed_sparse_assign_probs = tf.sparse.reorder(transposed_sparse_assign_probs)
#
#     transposed_sparse_adj = tf.SparseTensor(
#         indices=tf.cast(tf.transpose(edge_index, [1, 0]), dtype=tf.int64),
#         values=edge_weight,
#         dense_shape=[num_nodes, num_nodes]
#     )
#     transposed_sparse_adj = tf.sparse.reorder(transposed_sparse_adj)
#
#     # compute S'AS
#     # Since tf only support sparse @ dense, we compute S'AS as S' @ (S' @ dense(A'))'
#     pooled_adj = tf.sparse.sparse_dense_matmul(transposed_sparse_assign_probs, tf.sparse.to_dense(transposed_sparse_adj))
#     pooled_adj = tf.sparse.sparse_dense_matmul(transposed_sparse_assign_probs, tf.transpose(pooled_adj, [1, 0]))
#
#     pooled_edge_index, pooled_edge_weight = convert_dense_adj_to_edge(pooled_adj)
#
#     # pooled_edge_index, pooled_edge_weight = remove_self_loop_edge(pooled_edge_index, pooled_edge_weight)
#
#     if x is not None:
#         pooled_x = tf.sparse.sparse_dense_matmul(transposed_sparse_assign_probs, x)
#     else:
#         pooled_x = None
#     return pooled_x, pooled_edge_index, pooled_edge_weight


# def _cluster_pool_new(x, edge_index, edge_weight, assign_edge_index, assign_edge_weight, num_clusters, num_nodes=None):
#     """
#     Coarsen the input Graph based on cluster assignment of nodes and output pooled Graph.
#
#     :param x: Tensor, shape: [num_nodes, num_features], node features
#     :param edge_index: Tensor, shape: [2, num_edges], edge information
#     :param edge_weight: Tensor or None, shape: [num_edges]
#     :param assign_edge_index: Tensor, shape: [2, num_nodes], edge between clusters and nodes, where each edge
#         denotes a node belongs to a specific cluster.
#     :param assign_edge_weight: Tensor or None, shape: [num_nodes], the corresponding weight for assign_edge_index
#     :param num_clusters: Number of clusters.
#     :param num_nodes: Number of nodes, Optional, used for boosting performance.
#     :return: Pooled Graph: [pooled_x, pooled_edge_index, pooled_edge_weight]
#     """
#
#     # Manually passing in num_nodes, num_clusters, and num_graphs can boost the performance
#     if num_nodes is None:
#         if x is None:
#             raise Exception("Please provide num_nodes if x is None")
#         else:
#             num_nodes = tf.shape(x)[0]
#
#     # assign_row, assign_col = assign_edge_index[0], assign_edge_index[1]
#
#     # compute S'AS
#     S = SparseMatrix(assign_edge_index, assign_edge_weight, shape=[num_nodes, num_clusters])
#     ST = S.transpose()
#     A = SparseMatrix(edge_index, edge_weight, shape=[num_nodes, num_nodes])
#
#
#     pooled_adj = ST @ A @ S
#
#     if x is not None:
#         # S'X
#         pooled_x = ST @ x
#     else:
#         pooled_x = None
#
#
#     # transposed_sparse_assign_probs = tf.SparseTensor(
#     #     indices=tf.cast(tf.stack([assign_col, assign_row], axis=1), dtype=tf.int64),
#     #     values=assign_edge_weight,
#     #     dense_shape=[num_clusters, num_nodes]
#     # )
#     # transposed_sparse_assign_probs = tf.sparse.reorder(transposed_sparse_assign_probs)
#     #
#     # transposed_sparse_adj = tf.SparseTensor(
#     #     indices=tf.cast(tf.transpose(edge_index, [1, 0]), dtype=tf.int64),
#     #     values=edge_weight,
#     #     dense_shape=[num_nodes, num_nodes]
#     # )
#     # transposed_sparse_adj = tf.sparse.reorder(transposed_sparse_adj)
#     #
#     # # compute S'AS
#     # # Since tf only support sparse @ dense, we compute S'AS as S' @ (S' @ dense(A'))'
#     # pooled_adj = tf.sparse.sparse_dense_matmul(transposed_sparse_assign_probs, tf.sparse.to_dense(transposed_sparse_adj))
#     # pooled_adj = tf.sparse.sparse_dense_matmul(transposed_sparse_assign_probs, tf.transpose(pooled_adj, [1, 0]))
#
#     # pooled_edge_index, pooled_edge_weight = convert_dense_adj_to_edge(pooled_adj)
#
#     # if x is not None:
#     #     pooled_x = tf.sparse.sparse_dense_matmul(transposed_sparse_assign_probs, x)
#     # else:
#     #     pooled_x = None
#
#     return pooled_x, pooled_adj.edge_index, pooled_adj.edge_weight
#
#
#
# if version.parse(tf.__version__) < version.parse("2.5.0"):
#     cluster_pool = _cluster_pool_old
# else:
#     cluster_pool = _cluster_pool_new



# coding=utf-8
import tensorflow as tf
from tf_geometric.sparse.sparse_matrix import SparseMatrix


def gin_updater(x, reduced_neighbor_msg, eps):
    return x * (1.0 + eps) + reduced_neighbor_msg


def gin(x, edge_index, mlp_model, eps=0.0, training=None):
    """

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param mlp_model: A neural network (multi-layer perceptrons).
    :param eps: float, optional, (default: :obj:`0.`).
    :param training: Whether currently executing in training or inference mode.
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    # h = aggregate_neighbors(
    #     x, edge_index, None,
    #     identity_mapper,
    #     sum_reducer,
    #     identity_updater
    # )

    # h = gin_updater(x, h, eps)

    num_nodes = tf.shape(x)[0]
    sparse_adj = SparseMatrix(edge_index, edge_weight=None, shape=[num_nodes, num_nodes])

    neighbor_h = sparse_adj @ x
    h = x * (1.0 + eps) + neighbor_h
    h = mlp_model(h, training=training)

    return h



# coding=utf-8

import tensorflow as tf
from tf_geometric.sparse.sparse_matrix import SparseMatrix


# sparse_adj @ diagonal_matrix
def sparse_diag_matmul(sparse: SparseMatrix, diagonal):
    return sparse.matmul_diag(diagonal)


# self @ diagonal_matrix
def diag_sparse_matmul(diagonal, sparse: SparseMatrix):
    return sparse.rmatmul_diag(diagonal)


# element-wise sparse_adj addition
def add(a: SparseMatrix, b: SparseMatrix):
    return a + b


# element-wise sparse_adj subtraction
def subtract(a: SparseMatrix, b: SparseMatrix):
    return a - b


# element-wise maximum(a, b)
def maximum(a: SparseMatrix, b: SparseMatrix):
    return a.merge(b, merge_mode="max")


# element-wise minimum(a, b)
def minimum(a: SparseMatrix, b: SparseMatrix):
    return a.merge(b, merge_mode="min")


# Construct a SparseAdj from diagonals
def diags(diagonals):
    return SparseMatrix.from_diagonals(diagonals)


# Construct a SparseAdj with ones on diagonal
def eye(num_nodes):
    return SparseMatrix.eye(num_nodes)